from __future__ import annotations

import math
from abc import ABC
from typing import Optional, Union

from bolt11 import MilliSatoshi
from pydantic import BaseModel, Field, ValidationError, validator

from .exceptions import LnurlResponseException
from .types import (
    CallbackUrl,
    CiphertextBase64,
    InitializationVectorBase64,
    LightningInvoice,
    LightningNodeUri,
    Lnurl,
    LnurlPayMetadata,
    LnurlPaySuccessActionTag,
    LnurlResponseTag,
    LnurlStatus,
    Max144Str,
    Url,
)


class LnurlPayRouteHop(BaseModel):
    nodeId: str
    channelUpdate: str


class LnurlPaySuccessAction(BaseModel, ABC):
    tag: LnurlPaySuccessActionTag


class MessageAction(LnurlPaySuccessAction):
    tag: LnurlPaySuccessActionTag = LnurlPaySuccessActionTag.message
    message: Max144Str


class UrlAction(LnurlPaySuccessAction):
    tag: LnurlPaySuccessActionTag = LnurlPaySuccessActionTag.url
    url: Url
    description: Max144Str


# LUD-10: Add support for AES encrypted messages in payRequest.
class AesAction(LnurlPaySuccessAction):
    tag: LnurlPaySuccessActionTag = LnurlPaySuccessActionTag.aes
    description: Max144Str
    ciphertext: CiphertextBase64
    iv: InitializationVectorBase64


class LnurlResponseModel(BaseModel):

    class Config:
        use_enum_values = True
        extra = "forbid"

    def dict(self, **kwargs):
        kwargs["exclude_none"] = True
        return super().dict(**kwargs)

    def json(self, **kwargs):
        kwargs["exclude_none"] = True
        return super().json(**kwargs)

    @property
    def ok(self) -> bool:
        return True


class LnurlErrorResponse(LnurlResponseModel):
    status: LnurlStatus = LnurlStatus.error
    reason: str

    @property
    def error_msg(self) -> str:
        return self.reason

    @property
    def ok(self) -> bool:
        return False


class LnurlSuccessResponse(LnurlResponseModel):
    status: LnurlStatus = LnurlStatus.ok


# LUD-21: verify base spec.
class LnurlPayVerifyResponse(LnurlSuccessResponse):
    pr: LightningInvoice = Field(description="Payment request")
    settled: bool = Field(description="Settled status of the payment")
    preimage: Optional[str] = Field(default=None, description="Preimage of the payment")


# LUD-04: auth base spec.
class LnurlAuthResponse(LnurlResponseModel):
    tag: LnurlResponseTag = LnurlResponseTag.login
    callback: CallbackUrl
    k1: str


# LUD-2: channelRequest base spec.
class LnurlChannelResponse(LnurlResponseModel):
    tag: LnurlResponseTag = LnurlResponseTag.channelRequest
    uri: LightningNodeUri
    callback: CallbackUrl
    k1: str


# LUD-07: hostedChannelRequest base spec.
class LnurlHostedChannelResponse(LnurlResponseModel):
    tag: LnurlResponseTag = LnurlResponseTag.hostedChannelRequest
    uri: LightningNodeUri
    k1: str
    alias: Optional[str] = None


# LUD-18: Payer identity in payRequest protocol.
class LnurlPayResponsePayerDataOption(BaseModel):
    mandatory: bool


class LnurlPayResponsePayerDataOptionAuth(LnurlPayResponsePayerDataOption):
    k1: str


class LnurlPayResponsePayerDataExtra(BaseModel):
    name: str
    field: LnurlPayResponsePayerDataOption


class LnurlPayResponsePayerData(BaseModel):
    name: Optional[LnurlPayResponsePayerDataOption] = None
    pubkey: Optional[LnurlPayResponsePayerDataOption] = None
    identifier: Optional[LnurlPayResponsePayerDataOption] = None
    email: Optional[LnurlPayResponsePayerDataOption] = None
    auth: Optional[LnurlPayResponsePayerDataOptionAuth] = None
    extras: Optional[list[LnurlPayResponsePayerDataExtra]] = None


class LnurlPayerDataAuth(BaseModel):
    key: str
    k1: str
    sig: str


class LnurlPayerData(BaseModel):
    name: Optional[str] = None
    pubkey: Optional[str] = None
    identifier: Optional[str] = None
    email: Optional[str] = None
    auth: Optional[LnurlPayerDataAuth] = None
    extras: Optional[dict] = None


class LnurlPayResponse(LnurlResponseModel):
    tag: LnurlResponseTag = LnurlResponseTag.payRequest
    callback: CallbackUrl
    minSendable: MilliSatoshi = Field(gt=0)
    maxSendable: MilliSatoshi = Field(gt=0)
    metadata: LnurlPayMetadata
    # LUD-18: Payer identity in payRequest protocol.
    payerData: Optional[LnurlPayResponsePayerData] = None
    # Adds the optional comment_allowed field to the LnurlPayResponse
    # ref LUD-12: Comments in payRequest.
    commentAllowed: Optional[int] = None

    # NIP-57 Lightning Zaps
    allowsNostr: Optional[bool] = None
    nostrPubkey: Optional[str] = None

    @validator("maxSendable")
    def max_less_than_min(cls, value, values):  # noqa
        if "minSendable" in values and value < values["minSendable"]:
            raise ValueError("`maxSendable` cannot be less than `minSendable`.")
        return value

    @property
    def min_sats(self) -> int:
        return int(math.ceil(self.minSendable / 1000))

    @property
    def max_sats(self) -> int:
        return int(math.floor(self.maxSendable / 1000))

    def is_valid_amount(self, amount_msat: int) -> bool:
        return self.minSendable <= amount_msat <= self.maxSendable


class LnurlPayActionResponse(LnurlResponseModel):
    pr: LightningInvoice
    # LUD-9: successAction field for payRequest.
    successAction: Optional[Union[AesAction, MessageAction, UrlAction]] = None
    routes: list[list[LnurlPayRouteHop]] = []
    # LUD-11: Disposable and storeable payRequests.
    # If disposable is null, it should be interpreted as true.
    # so if SERVICE intends its LNURL links to be stored it must return disposable=False.
    disposable: Optional[bool] = Field(default=None, description="LUD-11: Disposable and storeable payRequests.")
    # LUD-21: verify base spec.
    verify: Optional[CallbackUrl] = Field(default=None, description="LUD-21: verify base spec.")


class LnurlWithdrawResponse(LnurlResponseModel):
    tag: LnurlResponseTag = LnurlResponseTag.withdrawRequest
    callback: CallbackUrl
    k1: str
    minWithdrawable: MilliSatoshi = Field(ge=0)
    maxWithdrawable: MilliSatoshi = Field(ge=0)
    defaultDescription: str = ""
    # LUD-14: balanceCheck: reusable withdrawRequests
    balanceCheck: Optional[CallbackUrl] = None
    currentBalance: Optional[MilliSatoshi] = None
    # LUD-19: Pay link discoverable from withdraw link.
    payLink: Optional[str] = None

    @validator("payLink", pre=True)
    def paylink_must_be_lud17(cls, value: Optional[str] = None) -> str | None:
        if not value:
            return None
        lnurl = Lnurl(value)
        if lnurl.is_lud17 and lnurl.lud17_prefix == "lnurlp":
            return value
        raise ValueError("`payLink` must be a valid LUD17 URL (lnurlp://).")

    @validator("maxWithdrawable")
    def max_less_than_min(cls, value, values):
        if "minWithdrawable" in values and value < values["minWithdrawable"]:
            raise ValueError("`maxWithdrawable` cannot be less than `minWithdrawable`.")
        return value

    # LUD-08: Fast withdrawRequest.
    @property
    def fast_withdraw_query(self) -> str:
        return "&".join([f"{k}={v}" for k, v in self.dict().items()])

    @property
    def min_sats(self) -> int:
        return int(math.ceil(self.minWithdrawable / 1000))

    @property
    def max_sats(self) -> int:
        return int(math.floor(self.maxWithdrawable / 1000))

    def is_valid_amount(self, amount: int) -> bool:
        return self.minWithdrawable <= amount <= self.maxWithdrawable


def is_pay_action_response(data: dict) -> bool:
    return "pr" in data and "routes" in data


class LnurlResponse:

    @staticmethod
    def from_dict(data: dict) -> LnurlResponseModel:
        tag = data.get("tag")

        # drop status field from all responses with a tag
        # some services return `status` in responses with a tag
        if tag or is_pay_action_response(data):
            data.pop("status", None)

        try:
            if tag == "channelRequest":
                return LnurlChannelResponse(**data)
            elif tag == "hostedChannelRequest":
                return LnurlHostedChannelResponse(**data)
            elif tag == "payRequest":
                return LnurlPayResponse(**data)
            elif tag == "withdrawRequest":
                return LnurlWithdrawResponse(**data)
            elif is_pay_action_response(data):
                return LnurlPayActionResponse(**data)

        except ValidationError as exc:
            raise LnurlResponseException(str(exc)) from exc

        status = data.get("status")
        if status is None or status == "":
            raise LnurlResponseException("Expected Success or Error response. But no `status` given.")

        # some services return `status` in lowercase, but spec says upper
        status = status.upper()

        if status == "OK":
            return LnurlSuccessResponse(status=LnurlStatus.ok)

        if status == "ERROR":
            return LnurlErrorResponse(status=LnurlStatus.error, reason=data.get("reason", "Unknown error"))

        # if we reach here, it's an unknown response
        raise LnurlResponseException(f"Unknown response: {data}")



from lnurl import (
    lnurlauth_derive_linking_key,
    lnurlauth_derive_linking_key_sign_message,
    lnurlauth_derive_path,
    lnurlauth_master_key_from_seed,
    lnurlauth_message_to_sign,
    lnurlauth_signature,
    lnurlauth_verify,
)

# taken from LUD-04 signature check example
k1 = "e2af6254a8df433264fa23f67eb8188635d15ce883e8fc020989d5f82ae6f11e"
key = "02c3b844b8104f0c1b15c507774c9ba7fc609f58f343b9b149122e944dd20c9362"
sig = (
    "304402203767faf494f110b139293d9bab3c50e07b3bf33c463d4aa767256cd09132dc510"
    "2205821f8efacdb5c595b92ada255876d9201e126e2f31a140d44561cc1f7e9e43d"
)

# LUD-05 derive path example
domain_name = "site.com"
hashing_private_key = "7d417a6a5e9a6a4a879aeaba11a11838764c8fa2b959c242d43dea682b3e409b"
path_suffix = "m/138'/1588488367/2659270754/38110259/4136336762"

# LUD-13 phrase sha256
phrase = "4b8dac0e71a99c61d2c197e6932fab3ae3cf7900fc19076864aa11e80d83b4d9"


class TestHelpersLnurlauth:

    def test_verify(self):
        assert lnurlauth_verify(k1, key, sig) is True

    def test_verify_invalid(self):
        k1 = "0" * 32
        key = "0" * 33
        sig = "0" * 64
        assert lnurlauth_verify(k1, key, sig) is False

    def test_signature(self):
        linking_key, _ = lnurlauth_derive_linking_key(key, domain_name)
        _key, _sig = lnurlauth_signature(k1, linking_key)
        assert lnurlauth_verify(k1, _key, _sig) is True
        # invalid signature
        assert lnurlauth_verify(k1, key, _sig) is False

    def test_derive_path(self):
        _path = lnurlauth_derive_path(bytes.fromhex(hashing_private_key), domain_name)
        assert _path == path_suffix

    def test_master_key_from_seed(self):
        _master = lnurlauth_master_key_from_seed(key)
        assert _master
        assert _master.privkey

    def test_phrase_sha256(self):
        assert lnurlauth_message_to_sign().hex() == phrase

    def test_linking_key_signmessage(self):
        _ = lnurlauth_derive_linking_key_sign_message(domain_name, b"0" * 32)



import pytest
from pydantic import parse_obj_as

from lnurl import encode
from lnurl.models import LnurlWithdrawResponse

url = "https://lnbits.com/withdraw/api/v1/lnurl"
url2 = f"{url}?tag=withdrawRequest"
url3 = f"{url2}&k1={16  * '0'}"
url4 = f"{url3}&callback={url}&defaultDescription=default"
url5 = f"{url4}&minWithdrawable=1000&maxWithdrawable=1000000"


class TestFastWithdraw:
    @pytest.mark.parametrize(
        "url, expected",
        [
            (url, False),
            (url2, False),
            (url3, False),
            (url4, False),
            (url5, True),
        ],
    )
    def test_is_lnurl_fast_withdraw(self, url: str, expected: bool):
        lnurl = encode(url)
        assert lnurl.is_fast_withdraw == expected

    def test_set_lnurl_fast_withdraw(self):
        response = parse_obj_as(
            LnurlWithdrawResponse,
            {
                "tag": "withdrawRequest",
                "k1": "0" * 16,
                "minWithdrawable": 1000,
                "maxWithdrawable": 1000000,
                "defaultDescription": "default",
                "callback": url,
            },
        )
        lnurl = encode(f"{url}?{response.fast_withdraw_query}")
        assert lnurl.is_fast_withdraw is True



from base64 import b64decode, b64encode

import pytest
from Cryptodome import Random

from lnurl.helpers import aes_decrypt, aes_encrypt


class TestAesEncryption:
    @pytest.mark.parametrize("plaintext", ["dni was here", "short", "A" * 4096])
    def test_encrypt_decrypt(self, plaintext):
        preimage = Random.get_random_bytes(32)
        ciphertext, iv = aes_encrypt(preimage, plaintext)
        assert ciphertext != plaintext
        assert len(b64decode(ciphertext)) % 16 == 0
        assert len(iv) == 24
        assert len(b64decode(iv)) == 16
        assert aes_decrypt(preimage, ciphertext, iv) == plaintext
        with pytest.raises(ValueError):
            assert aes_decrypt(bytes(32), ciphertext, iv) == plaintext
        with pytest.raises(ValueError):
            assert aes_decrypt(preimage, ciphertext, b64encode(bytes(16)).decode()) == plaintext
        with pytest.raises(ValueError):
            assert aes_decrypt(preimage, b64encode(bytes(32)).decode(), iv) == plaintext

    def test_encrypt_empty(self):
        with pytest.raises(ValueError):
            _ = aes_encrypt(b64encode(bytes(32)), "")

    def test_encrypt_fails_too_small_key(self):
        with pytest.raises(ValueError):
            _ = aes_encrypt(b64encode(bytes(33)), "dni was here")

    def test_encrypt_fails_too_big_key(self):
        with pytest.raises(ValueError):
            _ = aes_encrypt(b64encode(bytes(31)), "dni was here")

    # TODO: interesting, why?
    def test_encrypt_fails_for_icons(self):
        icons = "lightning icons: ⚡⚡"
        with pytest.raises(Exception):
            _ = aes_encrypt(b64encode(bytes(32)), icons)



import pytest
from pydantic import ValidationError, parse_obj_as

from lnurl import (
    CallbackUrl,
    LightningInvoice,
    LightningNodeUri,
    LnAddress,
    LnAddressError,
    Lnurl,
    LnurlErrorResponse,
    LnurlPayerData,
    LnurlPayerDataAuth,
    LnurlPayMetadata,
    LnurlPayResponsePayerData,
    LnurlPayResponsePayerDataExtra,
    LnurlPayResponsePayerDataOption,
    LnurlPayResponsePayerDataOptionAuth,
)


class TestUrl:
    # https://github.com/pydantic/pydantic/discussions/2450
    @pytest.mark.parametrize(
        "hostport",
        ["service.io:443", "service.io:9000"],
    )
    def test_parameters(self, hostport):
        url = parse_obj_as(CallbackUrl, f"https://{hostport}/?q=3fc3645b439ce8e7&test=ok")
        assert url.host == "service.io"
        assert url == f"https://{hostport}/?q=3fc3645b439ce8e7&test=ok"
        assert url.query_params == {"q": "3fc3645b439ce8e7", "test": "ok"}

    @pytest.mark.parametrize(
        "url",
        [
            "https://service.io/?q=3fc3645b439ce8e7&test=ok",
            "https://[2001:db8:0:1]:80",
            "https://protonirockerxow.onion/",
            "http://protonirockerxow.onion/",
            "https://📙.la/⚡",  # https://emojipedia.org/high-voltage-sign/
            "https://xn--yt8h.la/%E2%9A%A1",
            "http://0.0.0.0",
            "http://127.0.0.1",
            "http://localhost",
        ],
    )
    def test_valid_callback(self, url):
        url = parse_obj_as(CallbackUrl, url)
        assert isinstance(url, CallbackUrl)

    @pytest.mark.parametrize(
        "url",
        [
            "http://service.io/?q=3fc3645b439ce8e7&test=ok",
            "http://[2001:db8:0:1]:80",
            f'https://service.io/?hash={"x" * 4096}',
            "https://1.1.1.1/\u0000",
            "http://xn--yt8h.la/%E2%9A%A1",
            "http://1.1.1.1",
            "lnurlp://service.io",
        ],
    )
    def test_invalid_data_callback(self, url):
        with pytest.raises(ValidationError):
            parse_obj_as(CallbackUrl, url)

    @pytest.mark.parametrize(
        "url",
        [
            "https://📙.la/⚡",  # https://emojipedia.org/high-voltage-sign/
            "https://xn--yt8h.la/%E2%9A%A1",
        ],
    )
    def test_strict_rfc3986(self, monkeypatch, url):
        monkeypatch.setenv("LNURL_STRICT_RFC3986", "1")
        with pytest.raises(ValidationError):
            parse_obj_as(CallbackUrl, url)


class TestLightningInvoice:
    @pytest.mark.parametrize(
        "bech32, hrp, prefix, amount, h",
        [
            (
                (
                    "lntb20m1pvjluezpp5qqqsyqcyq5rqwzqfqqqsyqcyq5rqwzqfqqqsyqcyq5rqwzqfqypqhp58yjmdan79s6qqdhdzgynm4zwqd"
                    "5d7xmw5fk98klysy043l2ahrqsfpp3x9et2e20v6pu37c5d9vax37wxq72un98k6vcx9fz94w0qf237cm2rqv9pmn5lnexfvf55"
                    "79slr4zq3u8kmczecytdx0xg9rwzngp7e6guwqpqlhssu04sucpnz4axcv2dstmknqq6jsk2l"
                ),
                "lntb20m",
                "lntb",
                20,
                "h",
            ),
        ],
    )
    def test_valid_invoice(self, bech32, hrp, prefix, amount, h):
        invoice = LightningInvoice(bech32)
        assert invoice == parse_obj_as(LightningInvoice, bech32)
        assert invoice.hrp == hrp
        # TODO: implement these properties
        # assert invoice.prefix == prefix
        # assert invoice.amount == amount


class TestLightningNode:
    def test_valid_node(self):
        node = parse_obj_as(LightningNodeUri, "node_key@ip_address:port_number")
        assert node.key == "node_key"
        assert node.ip == "ip_address"
        assert node.port == "port_number"

    @pytest.mark.parametrize("uri", ["https://service.io/node", "node_key@ip_address", "ip_address:port_number"])
    def test_invalid_node(self, uri):
        with pytest.raises(ValidationError):
            parse_obj_as(LightningNodeUri, uri)


class TestLnurl:
    @pytest.mark.parametrize(
        "lightning",
        [
            (
                "LNURL1DP68GURN8GHJ7UM9WFMXJCM99E5K7TELWY7NXENRXVMRGDTZXSENJCM98PJNWE3JX56NXCFK89JN2V3K"
                "XUCRSVTY8YMXGCMYXV6RQD3EXDSKVCTZV5CRGCN9XA3RQCMRVSCNWWRYVCYAE0UU"
            ),
            (
                "lightning:LNURL1DP68GURN8GHJ7UM9WFMXJCM99E5K7TELWY7NXENRXVMRGDTZXSENJCM98PJNWE3JX56NXCFK89JN2V3K"
                "XUCRSVTY8YMXGCMYXV6RQD3EXDSKVCTZV5CRGCN9XA3RQCMRVSCNWWRYVCYAE0UU"
            ),
            "https://service.io?a=1&b=2",
            "lnurlp://service.io?a=1&b=2",
            "lnurlp://service.io/lnurlp?a=1&b=2",
        ],
    )
    def test_valid_lnurl_and_bech32(self, lightning):
        lnurl = Lnurl(lightning)
        assert lnurl == parse_obj_as(Lnurl, lightning)
        if lnurl.is_lud17:
            assert lnurl.lud17 == lightning
            assert lnurl.lud17_prefix == lightning.split("://")[0]
            assert lnurl.is_lud17 is True
            assert lnurl.url == lightning.replace("lnurlp://", "https://")
            assert lnurl == lightning.replace("lnurlp://", "https://")
        else:
            assert lnurl.bech32 is not None
            assert lnurl.bech32.hrp == "lnurl"
            assert lnurl.bech32 == lnurl or lnurl.url == lnurl
            assert lnurl.lud17 is None
            assert lnurl.lud17_prefix is None
            assert lnurl.is_lud17 is False

        assert lnurl.is_login is False

    @pytest.mark.parametrize(
        "url",
        [
            "lnurlp://service.io?a=1&b=2",
            "lnurlc://service.io",
            "lnurlw://service.io",
            "keyauth://service.io",
        ],
    )
    def test_valid_lnurl_lud17(self, url: str):
        _lnurl = parse_obj_as(Lnurl, url)

        _prefix = url.split("://")[0]
        assert _lnurl.lud17 == url
        assert _lnurl.lud17_prefix == _prefix
        assert _lnurl.is_lud17 is True
        _url = url.replace("lnurlp://", "https://")
        _url = _url.replace("lnurlc://", "https://")
        _url = _url.replace("lnurlw://", "https://")
        _url = _url.replace("keyauth://", "https://")
        assert str(_lnurl.url) == _url
        assert str(_lnurl) == _url

    @pytest.mark.parametrize(
        "url",
        [
            "http://service.io",
            "lnurlx://service.io",
        ],
    )
    def test_invalid_lnurl(self, url: str):
        with pytest.raises(ValidationError):
            parse_obj_as(Lnurl, url)

    @pytest.mark.parametrize(
        "bech32",
        [
            "abcdef1qpzry9x8gf2tvdw0s3jn54khce6mua7lmqqqxw",
            "split1checkupstagehandshakeupstreamerranterredcaperred2y9e3w",
            "BC1QW508D6QEJXTDG4Y5R3ZARVARY0C5XW7KV8F3T4",
        ],
    )
    def test_decode_nolnurl(self, bech32):
        with pytest.raises(ValidationError):
            parse_obj_as(Lnurl, bech32)

    @pytest.mark.parametrize(
        "url",
        [
            "lnurlp://localhost",
            "http://localhost",
        ],
    )
    def test_insecure_lnurl(self, url: str):
        lnurl = parse_obj_as(Lnurl, url)
        assert lnurl.url.insecure is True
        assert lnurl.url.host == "localhost"
        assert lnurl.url.startswith("http://")


class TestLnurlPayMetadata:
    @pytest.mark.parametrize(
        "metadata, image_type",
        [
            ('[["text/plain", "main text"]]', None),
            ('[["text/plain", "main text"], ["image/jpeg;base64", "base64encodedimage"]]', "jpeg"),
            ('[["text/plain", "main text"], ["image/png;base64", "base64encodedimage"]]', "png"),
            ('[["text/plain", "main text"], ["text/indentifier", "alan@lnbits.com"], ["text/tag", "tag"]]', None),
        ],
    )
    def test_valid(self, metadata, image_type):
        m = parse_obj_as(LnurlPayMetadata, metadata)
        assert m.text == "main text"

        if m.images:
            assert len(m.images) == 1
            assert dict(m.images)[f"image/{image_type};base64"] == "base64encodedimage"

    @pytest.mark.parametrize(
        "metadata",
        [
            "[]",
            '["text""plain"]',
            '[["text", "plain"]]',
            '[["text", "plain", "plane"]]',
            '[["text/plain", "main text"], ["text/plain", "two is too much"]]',
            '[["image/jpeg;base64", "base64encodedimage"]]',
        ],
    )
    def test_invalid_data(self, metadata):
        with pytest.raises(ValidationError):
            parse_obj_as(LnurlPayMetadata, metadata)

    @pytest.mark.parametrize(
        "lnaddress",
        [
            "donate@legend.lnbits.com",
        ],
    )
    def test_valid_lnaddress(self, lnaddress):
        lnaddress = LnAddress(lnaddress)
        assert isinstance(lnaddress.url, CallbackUrl)
        assert lnaddress.tag is None

    @pytest.mark.parametrize(
        "lnaddress",
        [
            "donate+lud16tag@legend.lnbits.com",
        ],
    )
    def test_valid_lnaddress_with_tag(self, lnaddress):
        lnaddress = LnAddress(lnaddress)
        assert isinstance(lnaddress.url, CallbackUrl)
        assert lnaddress.tag == "lud16tag"

    @pytest.mark.parametrize(
        "lnaddress",
        [
            "legend.lnbits.com",
            "donate@donate@legend.lnbits.com",
            "HELLO@lnbits.com",
        ],
    )
    def test_invalid_lnaddress(self, lnaddress):
        with pytest.raises(LnAddressError):
            lnaddress = LnAddress(lnaddress)


class TestPayerData:

    def test_valid_pay_response_payer_data(self):
        data = {
            "name": {"mandatory": True},
            "pubkey": {"mandatory": True},
            "auth": {"mandatory": True, "k1": "0" * 32},
            "extras": [
                {
                    "name": "extra_field",
                    "field": {"mandatory": True},
                },
            ],
        }
        payer_data = parse_obj_as(LnurlPayResponsePayerData, data)
        assert payer_data.name is not None
        assert payer_data.name.mandatory is True
        assert payer_data.pubkey is not None
        assert payer_data.pubkey.mandatory is True
        assert payer_data.auth is not None
        assert payer_data.auth.mandatory is True
        assert payer_data.auth.k1 == "0" * 32
        assert payer_data.extras is not None
        assert len(payer_data.extras) == 1
        assert payer_data.extras[0].name == "extra_field"
        assert payer_data.extras[0].field is not None

    def test_valid_pay_response_payer_data_models(self):
        data_option = LnurlPayResponsePayerDataOption(
            mandatory=True,
        )
        payer_data = LnurlPayResponsePayerData(
            name=data_option,
            pubkey=data_option,
            auth=LnurlPayResponsePayerDataOptionAuth(
                mandatory=True,
                k1="0" * 32,
            ),
            extras=[
                LnurlPayResponsePayerDataExtra(
                    name="extra_field",
                    field=LnurlPayResponsePayerDataOption(
                        mandatory=True,
                    ),
                ),
            ],
        )
        assert payer_data.name is not None
        assert payer_data.name.mandatory is True
        assert payer_data.pubkey is not None
        assert payer_data.pubkey.mandatory is True
        assert payer_data.auth is not None
        assert payer_data.auth.mandatory is True
        assert payer_data.auth.k1 == "0" * 32
        assert payer_data.extras is not None
        assert len(payer_data.extras) == 1
        assert payer_data.extras[0].name == "extra_field"
        assert payer_data.extras[0].field is not None

    def test_valid_payer_data(self):
        data = {
            "name": "John Doe",
            "pubkey": "03a3xxxxxxxxxxxx",
            "auth": {
                "key": "key",
                "k1": "0" * 32,
                "sig": "0" * 64,
            },
        }
        payer_data = parse_obj_as(LnurlPayerData, data)
        assert payer_data.name == "John Doe"
        assert payer_data.pubkey == "03a3xxxxxxxxxxxx"
        assert payer_data.auth is not None
        assert payer_data.auth.key == "key"
        assert payer_data.auth.k1 == "0" * 32
        assert payer_data.auth.sig == "0" * 64

    def test_valid_pay_response_payer_model(self):
        data = LnurlPayerData(
            name="John Doe",
            pubkey="03a3xxxxxxxxxxxx",
            auth=LnurlPayerDataAuth(
                key="key",
                k1="0" * 32,
                sig="0" * 64,
            ),
            extras={
                "extra_field": "extra_value",
            },
        )
        assert data.name == "John Doe"
        assert data.pubkey == "03a3xxxxxxxxxxxx"
        assert data.auth is not None
        assert data.auth.key == "key"
        assert data.auth.k1 == "0" * 32
        assert data.auth.sig == "0" * 64
        assert data.extras is not None
        assert data.extras["extra_field"] == "extra_value"


class TestLnurlErrorResponse:
    def test_error_res_details(self):
        res = LnurlErrorResponse(reason="detail")
        _dict = res.dict()
        assert "status" in _dict
        assert _dict["status"] == "ERROR"
        assert "reason" in _dict
        assert _dict["reason"] == "detail"
        assert res.json() == '{"status": "ERROR", "reason": "detail"}'



# backward compatibility, MilliSatoshi is now imported from bolt11
from bolt11 import MilliSatoshi

from .core import decode, encode, execute, execute_login, execute_pay_request, execute_withdraw, get, handle
from .exceptions import (
    InvalidLnurl,
    InvalidLnurlPayMetadata,
    InvalidUrl,
    LnAddressError,
    LnurlException,
    LnurlResponseException,
)
from .helpers import (
    LUD13_PHRASE,
    aes_decrypt,
    aes_encrypt,
    lnurlauth_derive_linking_key,
    lnurlauth_derive_linking_key_sign_message,
    lnurlauth_derive_path,
    lnurlauth_master_key_from_seed,
    lnurlauth_message_to_sign,
    lnurlauth_signature,
    lnurlauth_verify,
    url_decode,
    url_encode,
)
from .models import (
    AesAction,
    LnurlAuthResponse,
    LnurlChannelResponse,
    LnurlErrorResponse,
    LnurlHostedChannelResponse,
    LnurlPayActionResponse,
    LnurlPayerData,
    LnurlPayerDataAuth,
    LnurlPayResponse,
    LnurlPayResponsePayerData,
    LnurlPayResponsePayerDataExtra,
    LnurlPayResponsePayerDataOption,
    LnurlPayResponsePayerDataOptionAuth,
    LnurlPayRouteHop,
    LnurlPaySuccessAction,
    LnurlPayVerifyResponse,
    LnurlResponse,
    LnurlResponseModel,
    LnurlSuccessResponse,
    LnurlWithdrawResponse,
    MessageAction,
    UrlAction,
)
from .types import (
    Bech32,
    CallbackUrl,
    CiphertextBase64,
    InitializationVectorBase64,
    LightningInvoice,
    LightningNodeUri,
    LnAddress,
    Lnurl,
    LnurlAuthActions,
    LnurlPayMetadata,
    LnurlPaySuccessActionTag,
    LnurlResponseTag,
    LnurlStatus,
    Max144Str,
    Url,
)

__all__ = [
    "aes_decrypt",
    "aes_encrypt",
    "url_encode",
    "url_decode",
    "decode",
    "encode",
    "execute",
    "execute_login",
    "execute_pay_request",
    "execute_withdraw",
    "get",
    "handle",
    "Lnurl",
    "LnurlAuthActions",
    "LnurlAuthResponse",
    "LnurlChannelResponse",
    "LnurlErrorResponse",
    "LnurlHostedChannelResponse",
    "LnurlPayActionResponse",
    "LnurlPayResponse",
    "LnurlPayMetadata",
    "LnurlPaySuccessAction",
    "LnurlPaySuccessActionTag",
    "LnurlPayVerifyResponse",
    "LnurlResponse",
    "LnurlResponseModel",
    "LnurlResponseTag",
    "LnurlStatus",
    "LnurlSuccessResponse",
    "LnurlWithdrawResponse",
    "MilliSatoshi",
    "CallbackUrl",
    "Url",
    "LightningNodeUri",
    "LnAddress",
    "LightningInvoice",
    "Bech32",
    "LnurlPayRouteHop",
    "MessageAction",
    "UrlAction",
    "AesAction",
    "InitializationVectorBase64",
    "CiphertextBase64",
    "lnurlauth_signature",
    "lnurlauth_verify",
    "lnurlauth_derive_linking_key",
    "lnurlauth_master_key_from_seed",
    "lnurlauth_derive_path",
    "lnurlauth_message_to_sign",
    "lnurlauth_derive_linking_key_sign_message",
    "LUD13_PHRASE",
    "LnurlException",
    "LnAddressError",
    "InvalidLnurl",
    "InvalidLnurlPayMetadata",
    "InvalidUrl",
    "LnurlResponseException",
    "LnurlPayerData",
    "LnurlPayerDataAuth",
    "LnurlPayResponsePayerData",
    "LnurlPayResponsePayerDataOption",
    "LnurlPayResponsePayerDataOptionAuth",
    "LnurlPayResponsePayerDataExtra",
    "Max144Str",
]



import httpx
import pytest

from lnurl.core import decode, encode, execute_login, execute_pay_request, get, handle
from lnurl.exceptions import InvalidLnurl, InvalidUrl, LnurlResponseException
from lnurl.models import (
    LnurlAuthResponse,
    LnurlPayActionResponse,
    LnurlPayResponse,
    LnurlPaySuccessAction,
    LnurlSuccessResponse,
    LnurlWithdrawResponse,
)
from lnurl.types import Lnurl, Url


class TestDecode:
    @pytest.mark.parametrize(
        "bech32, url",
        [
            (
                (
                    "LNURL1DP68GURN8GHJ7UM9WFMXJCM99E5K7TELWY7NXENRXVMRGDTZXSENJCM98PJNWE3JX56NXCFK89JN2V3K"
                    "XUCRSVTY8YMXGCMYXV6RQD3EXDSKVCTZV5CRGCN9XA3RQCMRVSCNWWRYVCYAE0UU"
                ),
                "https://service.io/?q=3fc3645b439ce8e7f2553a69e5267081d96dcd340693afabe04be7b0ccd178df",
            )
        ],
    )
    def test_decode(self, bech32, url):
        decoded = decode(bech32)
        assert isinstance(decoded, Lnurl)
        decoded_url = decoded.url
        assert isinstance(decoded_url, Url)
        assert decoded_url == str(decoded_url) == url
        assert decoded_url.host == "service.io"

    @pytest.mark.parametrize(
        "bech32",
        [
            "abcdef1qpzry9x8gf2tvdw0s3jn54khce6mua7lmqqqxw",
            "split1checkupstagehandshakeupstreamerranterredcaperred2y9e3w",
            "BC1QW508D6QEJXTDG4Y5R3ZARVARY0C5XW7KV8F3T4",
        ],
    )
    def test_decode_nolnurl(self, bech32):
        with pytest.raises(InvalidLnurl):
            decode(bech32)


class TestEncode:
    @pytest.mark.parametrize(
        "bech32, url",
        [
            (
                (
                    "LNURL1DP68GURN8GHJ7UM9WFMXJCM99E5K7TELWY7NXENRXVMRGDTZXSENJCM98PJNWE3JX56NXCFK89JN2V3K"
                    "XUCRSVTY8YMXGCMYXV6RQD3EXDSKVCTZV5CRGCN9XA3RQCMRVSCNWWRYVCYAE0UU"
                ),
                "https://service.io/?q=3fc3645b439ce8e7f2553a69e5267081d96dcd340693afabe04be7b0ccd178df",
            )
        ],
    )
    def test_encode(self, bech32, url):
        lnurl = encode(url)
        assert isinstance(lnurl, Lnurl)
        assert lnurl.bech32 == bech32
        assert lnurl.url.host == "service.io"

    @pytest.mark.parametrize("url", ["http://service.io/"])
    def test_encode_nohttps(self, url):
        with pytest.raises(InvalidUrl):
            encode(url)


class TestHandle:
    """Responses from the LNURL: https://demo.lnbits.com/"""

    @pytest.mark.xfail(reason="demo.lnbits.com is down")
    @pytest.mark.asyncio
    @pytest.mark.parametrize(
        "bech32",
        [
            "LNURL1DP68GURN8GHJ7MR9VAJKUEPWD3HXY6T5WVHXXMMD9AMKJARGV3EXZAE0V9CXJTMKXYH"
            "KCMN4WFKZ7MJT2C6X2NRK0PDRYJNGWVU9WDN2G4V8XK2VSZA2RC"
        ],
    )
    async def test_handle_withdraw(self, bech32):
        res = await handle(bech32)
        assert isinstance(res, LnurlWithdrawResponse)
        assert res.tag == "withdrawRequest"
        assert res.callback.host == "demo.lnbits.com"
        assert res.default_description == "sample withdraw"
        assert res.max_withdrawable >= res.min_withdrawable

    @pytest.mark.asyncio
    @pytest.mark.parametrize("bech32", ["BC1QW508D6QEJXTDG4Y5R3ZARVARY0C5XW7KV8F3T4"])
    async def test_handle_nolnurl(self, bech32):
        with pytest.raises(InvalidLnurl):
            await handle(bech32)

    @pytest.mark.asyncio
    @pytest.mark.parametrize("url", ["https://lnurl.thisshouldfail.io/"])
    async def test_get_requests_error(self, url):
        with pytest.raises(LnurlResponseException):
            await get(url)


class TestPayFlow:
    """Full LNURL-pay flow interacting with https://demo.lnbits.com/"""

    @pytest.mark.xfail(reason="demo.lnbits.com is down")
    @pytest.mark.asyncio
    @pytest.mark.parametrize(
        "bech32, amount",
        [
            (
                "LNURL1DP68GURN8GHJ7MR9VAJKUEPWD3HXY6T5WVHXXMMD9AKXUATJD3CZ7JN9F4EHQJQC25ZZY",
                "1000",
            ),
            ("donate@demo.lnbits.com", "100000"),
        ],
    )
    async def test_pay_flow(self, bech32: str, amount: str):
        res = await handle(bech32)
        assert isinstance(res, LnurlPayResponse)
        assert res.tag == "payRequest"
        assert res.callback.host == "demo.lnbits.com"
        assert len(res.metadata.list()) >= 1
        assert res.metadata.text != ""

        res2 = await execute_pay_request(res, amount)
        assert isinstance(res2, LnurlPayActionResponse)
        assert res2.success_action is None or isinstance(res2.success_action, LnurlPaySuccessAction)


class TestLoginFlow:
    """Full LNURL-login flow interacting with https://lnmarkets.com/"""

    @pytest.mark.asyncio
    @pytest.mark.xfail(reason="need online lnurl auth server to test this flow")
    @pytest.mark.parametrize(
        "url",
        [
            "https://api.lnmarkets.com/trpc/lnurl.auth.new",
        ],
    )
    async def test_login_flow(self, url: str):
        async with httpx.AsyncClient() as client:
            init = await client.get(url)
            init.raise_for_status()
            bech32 = init.json()["result"]["data"]["json"]["lnurl"]

            res = await handle(bech32)
            assert isinstance(res, LnurlAuthResponse)
            assert res.tag == "login"
            assert res.callback.host == "api.lnmarkets.com"

            res2 = await execute_login(res, "my-secret")
            assert isinstance(res2, LnurlSuccessResponse)



import json
from base64 import b64encode

import pytest

from lnurl import (
    AesAction,
    LnurlErrorResponse,
    LnurlPayActionResponse,
    LnurlPayResponse,
    LnurlPaySuccessAction,
    LnurlPaySuccessActionTag,
    LnurlResponse,
    LnurlSuccessResponse,
    LnurlWithdrawResponse,
)
from lnurl.exceptions import LnurlResponseException


class TestLnurlResponse:
    pay_res = json.loads(
        r'{"tag":"payRequest","metadata":"[[\"text/plain\",\"lorem ipsum blah blah\"]]",'
        '"callback":"https://lnurl.bigsun.xyz/lnurl-pay/callback/","maxSendable":300980,'
        '"minSendable":100980}'
    )  # noqa
    pay_res_invalid = json.loads(r'{"tag":"payRequest","metadata":"[\"text\"\"plain\"]"}')
    withdraw_res = json.loads(
        '{"tag":"withdrawRequest","k1":"c67a8aa61f7c6cd457058916356ca80f5bfd00fa78ac2c1b3157391c2e9787de",'
        '"callback":"https://lnurl.bigsun.xyz/lnurl-withdraw/callback/?param1=1&param2=2",'
        '"maxWithdrawable":478980,"minWithdrawable":478980,"defaultDescription":"sample withdraw"}'
    )  # noqa
    pay_res_action_aes = {
        "pr": (
            "lnbc1pvjluezsp5zyg3zyg3zyg3zyg3zyg3zyg3zyg3zyg3zyg3zyg3zyg3zyg3zygspp5qqqsyqcyq5rqwzqfqqq"
            "syqcyq5rqwzqfqqqsyqcyq5rqwzqfqypqdpl2pkx2ctnv5sxxmmwwd5kgetjypeh2ursdae8g6twvus8g6rfwvs8q"
            "un0dfjkxaq9qrsgq357wnc5r2ueh7ck6q93dj32dlqnls087fxdwk8qakdyafkq3yap9us6v52vjjsrvywa6rt52c"
            "m9r9zqt8r2t7mlcwspyetp5h2tztugp9lfyql"
        ),
        "routes": [],
        "successAction": {
            "tag": "aes",
            "description": "your will receive a secret message",
            "iv": b64encode(bytes(16)),
            "ciphertext": b64encode(bytes(32)),
        },
    }

    def test_error(self):
        res = LnurlResponse.from_dict({"status": "error", "reason": "error details..."})
        assert isinstance(res, LnurlErrorResponse)
        assert not res.ok
        assert res.error_msg == "error details..."

    def test_success(self):
        res = LnurlResponse.from_dict({"status": "OK"})
        assert isinstance(res, LnurlSuccessResponse)
        assert res.ok

    def test_unknown(self):
        with pytest.raises(LnurlResponseException):
            LnurlResponse.from_dict({"status": "unknown"})

    def test_pay(self):
        res = LnurlResponse.from_dict(self.pay_res)
        assert isinstance(res, LnurlPayResponse)
        assert res.ok
        assert res.max_sats == 300
        assert res.min_sats == 101
        assert res.metadata == '[["text/plain","lorem ipsum blah blah"]]'
        assert res.metadata.list() == [("text/plain", "lorem ipsum blah blah")]
        assert not res.metadata.images
        assert res.metadata.text == "lorem ipsum blah blah"
        assert res.metadata.h == "d824d0ea606c5a9665279c31cf185528a8df2875ea93f1f75e501e354b33e90a"

    def test_pay_invalid_metadata(self):
        with pytest.raises(LnurlResponseException):
            LnurlResponse.from_dict(self.pay_res_invalid)

    # LUD-10
    def test_pay_action_aes(self):
        res = LnurlResponse.from_dict(self.pay_res_action_aes)
        assert isinstance(res, LnurlPayActionResponse)
        assert isinstance(res.successAction, AesAction)
        assert isinstance(res.successAction, LnurlPaySuccessAction)
        assert res.ok
        assert res.successAction
        assert res.successAction.tag == LnurlPaySuccessActionTag.aes
        assert res.successAction.description == "your will receive a secret message"
        assert len(res.successAction.iv) == 24
        assert len(res.successAction.ciphertext) == 44

    def test_withdraw(self):
        res = LnurlResponse.from_dict(self.withdraw_res)
        assert isinstance(res, LnurlWithdrawResponse)
        assert res.ok
        assert res.maxWithdrawable == 478980
        assert res.max_sats == 478
        assert res.min_sats == 479



from json import JSONDecodeError
from typing import Any, Optional

import httpx
from bolt11 import Bolt11Exception, MilliSatoshi
from bolt11 import decode as bolt11_decode
from pydantic import ValidationError, parse_obj_as

from .exceptions import InvalidLnurl, InvalidUrl, LnurlResponseException
from .helpers import (
    lnurlauth_derive_linking_key,
    lnurlauth_derive_linking_key_sign_message,
    lnurlauth_signature,
    url_encode,
)
from .models import (
    LnurlAuthResponse,
    LnurlErrorResponse,
    LnurlPayActionResponse,
    LnurlPayResponse,
    LnurlResponse,
    LnurlResponseModel,
    LnurlSuccessResponse,
    LnurlWithdrawResponse,
)
from .types import CallbackUrl, LnAddress, Lnurl

USER_AGENT = "lnbits/lnurl"
TIMEOUT = 5


def decode(lnurl: str) -> Lnurl:
    try:
        return Lnurl(lnurl)
    except (ValidationError, ValueError):
        raise InvalidLnurl


def encode(url: str) -> Lnurl:
    try:
        return Lnurl(url_encode(url))
    except (ValidationError, ValueError):
        raise InvalidUrl


async def get(
    url: str,
    *,
    response_class: Optional[Any] = None,
    user_agent: Optional[str] = None,
    timeout: Optional[int] = None,
) -> LnurlResponseModel:
    headers = {"User-Agent": user_agent or USER_AGENT}
    async with httpx.AsyncClient(headers=headers, follow_redirects=True) as client:
        try:
            res = await client.get(url, timeout=timeout or TIMEOUT)
            res.raise_for_status()
        except Exception as exc:
            raise LnurlResponseException(str(exc)) from exc

        try:
            _json = res.json()
        except JSONDecodeError as exc:
            raise LnurlResponseException(f"Invalid JSON response from {url}") from exc

        if response_class:
            if not issubclass(response_class, LnurlResponseModel):
                raise LnurlResponseException("response_class must be a subclass of LnurlResponseModel")
            return response_class(**_json)

        return LnurlResponse.from_dict(_json)


async def handle(
    lnurl: str,
    response_class: Optional[LnurlResponseModel] = None,
    user_agent: Optional[str] = None,
    timeout: Optional[int] = None,
) -> LnurlResponseModel:
    try:
        if "@" in lnurl:
            lnaddress = LnAddress(lnurl)
            return await get(lnaddress.url, response_class=response_class, user_agent=user_agent, timeout=timeout)
        lnurl = Lnurl(lnurl)
    except (ValidationError, ValueError):
        raise InvalidLnurl

    if lnurl.is_login:
        callback_url = parse_obj_as(CallbackUrl, lnurl.url)
        return LnurlAuthResponse(callback=callback_url, k1=lnurl.url.query_params["k1"])

    return await get(lnurl.url, response_class=response_class, user_agent=user_agent, timeout=timeout)


async def execute(
    bech32_or_address: str,
    value: str,
    user_agent: Optional[str] = None,
    timeout: Optional[int] = None,
) -> LnurlResponseModel:
    try:
        res = await handle(bech32_or_address, user_agent=user_agent, timeout=timeout)
    except Exception as exc:
        raise LnurlResponseException(str(exc))

    if isinstance(res, LnurlPayResponse) and res.tag == "payRequest":
        return await execute_pay_request(res, int(value), user_agent=user_agent, timeout=timeout)
    elif isinstance(res, LnurlAuthResponse) and res.tag == "login":
        return await execute_login(res, value, user_agent=user_agent, timeout=timeout)
    elif isinstance(res, LnurlWithdrawResponse) and res.tag == "withdrawRequest":
        return await execute_withdraw(res, value, user_agent=user_agent, timeout=timeout)

    raise LnurlResponseException("tag not implemented")


async def execute_pay_request(
    res: LnurlPayResponse,
    msat: int,
    comment: Optional[str] = None,
    user_agent: Optional[str] = None,
    timeout: Optional[int] = None,
) -> LnurlPayActionResponse:
    if not res.minSendable <= MilliSatoshi(msat) <= res.maxSendable:
        raise LnurlResponseException(f"Amount {msat} not in range {res.minSendable} - {res.maxSendable}")

    params: dict[str, str | int] = {"amount": msat}

    if res.commentAllowed and comment:
        if len(comment) > res.commentAllowed:
            raise LnurlResponseException(f"Comment length {len(comment)} exceeds allowed length {res.commentAllowed}")
        params["comment"] = comment

    try:
        headers = {"User-Agent": user_agent or USER_AGENT}
        async with httpx.AsyncClient(headers=headers, follow_redirects=True) as client:
            res2 = await client.get(
                url=res.callback,
                params=params,
                timeout=timeout or TIMEOUT,
            )
            res2.raise_for_status()
            pay_res = LnurlResponse.from_dict(res2.json())
            if isinstance(pay_res, LnurlErrorResponse):
                raise LnurlResponseException(pay_res.reason)
            if not isinstance(pay_res, LnurlPayActionResponse):
                raise LnurlResponseException(f"Expected LnurlPayActionResponse, got {type(pay_res)}")
            invoice = bolt11_decode(pay_res.pr)
            if invoice.amount_msat != int(msat):
                raise LnurlResponseException(
                    f"{res.callback.host} returned an invalid invoice."
                    f"Excepted `{msat}` msat, got `{invoice.amount_msat}`."
                )
            return pay_res
    except Exception as exc:
        raise LnurlResponseException(str(exc))


async def execute_login(
    res: LnurlAuthResponse,
    seed: str | None = None,
    signed_message: str | None = None,
    user_agent: Optional[str] = None,
    timeout: Optional[int] = None,
) -> LnurlResponseModel:
    if not res.callback:
        raise LnurlResponseException("LNURLauth callback does not exist")
    host = res.callback.host
    if not host:
        raise LnurlResponseException("Invalid host in LNURLauth callback")
    if seed:
        linking_key, _ = lnurlauth_derive_linking_key(seed=seed, domain=host)
    elif signed_message:
        linking_key, _ = lnurlauth_derive_linking_key_sign_message(domain=host, sig=signed_message.encode())
    else:
        raise LnurlResponseException("Seed or signed_message is required for LNURLauth")
    try:
        key, sig = lnurlauth_signature(res.k1, linking_key=linking_key)
        headers = {"User-Agent": user_agent or USER_AGENT}
        async with httpx.AsyncClient(headers=headers, follow_redirects=True) as client:
            res2 = await client.get(
                url=res.callback,
                params={
                    "key": key,
                    "sig": sig,
                },
                timeout=timeout or TIMEOUT,
            )
            res2.raise_for_status()
            return LnurlResponse.from_dict(res2.json())
    except Exception as e:
        raise LnurlResponseException(str(e))


async def execute_withdraw(
    res: LnurlWithdrawResponse,
    pr: str,
    user_agent: Optional[str] = None,
    timeout: Optional[int] = None,
) -> LnurlSuccessResponse:
    try:
        invoice = bolt11_decode(pr)
    except Bolt11Exception as exc:
        raise LnurlResponseException(str(exc))
    # if invoice does not have amount use the min withdrawable amount
    amount = invoice.amount_msat or res.minWithdrawable
    if not res.minWithdrawable <= MilliSatoshi(amount) <= res.maxWithdrawable:
        raise LnurlResponseException(f"Amount {amount} not in range {res.minWithdrawable} - {res.maxWithdrawable}")
    try:
        headers = {"User-Agent": user_agent or USER_AGENT}
        async with httpx.AsyncClient(headers=headers, follow_redirects=True) as client:
            res2 = await client.get(
                url=res.callback,
                params={
                    "k1": res.k1,
                    "pr": pr,
                },
                timeout=timeout or TIMEOUT,
            )
            res2.raise_for_status()
            withdraw_res = LnurlResponse.from_dict(res2.json())
            if isinstance(withdraw_res, LnurlErrorResponse):
                raise LnurlResponseException(withdraw_res.reason)
            if not isinstance(withdraw_res, LnurlSuccessResponse):
                raise LnurlResponseException(f"Expected LnurlSuccessResponse, got {type(withdraw_res)}")
            return withdraw_res
    except Exception as exc:
        raise LnurlResponseException(str(exc))



import json

import pytest
from pydantic import ValidationError, parse_obj_as

from lnurl import CallbackUrl, Lnurl, LnurlPayMetadata, MilliSatoshi, encode
from lnurl.models import (
    LnurlChannelResponse,
    LnurlErrorResponse,
    LnurlHostedChannelResponse,
    LnurlPayResponse,
    LnurlSuccessResponse,
    LnurlWithdrawResponse,
)


class TestLnurlErrorResponse:
    def test_response(self):
        res = LnurlErrorResponse(reason="blah blah blah")
        assert res.ok is False
        assert res.error_msg == "blah blah blah"
        assert res.json() == '{"status": "ERROR", "reason": "blah blah blah"}'
        assert res.dict() == {"status": "ERROR", "reason": "blah blah blah"}


class TestLnurlSuccessResponse:
    def test_success_response(self):
        res = LnurlSuccessResponse()
        assert res.ok
        assert res.json() == '{"status": "OK"}'
        assert res.dict() == {"status": "OK"}


class TestLnurlChannelResponse:
    @pytest.mark.parametrize(
        "d", [{"uri": "node_key@ip_address:port_number", "callback": "https://service.io/channel", "k1": "c3RyaW5n"}]
    )
    def test_channel_response(self, d):
        res = LnurlChannelResponse(**d)
        assert res.ok
        assert res.dict() == {**{"tag": "channelRequest"}, **d}

    @pytest.mark.parametrize(
        "d",
        [
            {"uri": "invalid", "callback": "https://service.io/channel", "k1": "c3RyaW5n"},
            {"uri": "node_key@ip_address:port_number", "callback": "invalid", "k1": "c3RyaW5n"},
            {"uri": "node_key@ip_address:port_number", "callback": "https://service.io/channel", "k1": None},
        ],
    )
    def test_invalid_data(self, d):
        with pytest.raises(ValidationError):
            LnurlChannelResponse(**d)


class TestLnurlHostedChannelResponse:
    @pytest.mark.parametrize("d", [{"uri": "node_key@ip_address:port_number", "k1": "c3RyaW5n"}])
    def test_channel_response(self, d):
        res = LnurlHostedChannelResponse(**d)
        assert res.ok
        assert res.dict() == {**{"tag": "hostedChannelRequest"}, **d}

    @pytest.mark.parametrize(
        "d", [{"uri": "invalid", "k1": "c3RyaW5n"}, {"uri": "node_key@ip_address:port_number", "k1": None}]
    )
    def test_invalid_data(self, d):
        with pytest.raises(ValidationError):
            LnurlHostedChannelResponse(**d)


metadata = '[["text/plain","lorem ipsum blah blah"]]'


class TestLnurlPayResponse:
    @pytest.mark.parametrize(
        "callback, min_sendable, max_sendable, metadata",
        [
            (
                "https://service.io/pay",
                1000,
                2000,
                metadata,
            ),
        ],
    )
    def test_success_response(self, callback: str, min_sendable: int, max_sendable: int, metadata: str):
        callback_url = parse_obj_as(CallbackUrl, callback)
        data = parse_obj_as(LnurlPayMetadata, metadata)
        res = LnurlPayResponse(
            callback=callback_url,
            minSendable=MilliSatoshi(min_sendable),
            maxSendable=MilliSatoshi(max_sendable),
            metadata=data,
        )
        assert res.ok
        assert (
            res.json() == res.json() == '{"tag": "payRequest", "callback": "https://service.io/pay", '
            f'"minSendable": 1000, "maxSendable": 2000, "metadata": {json.dumps(metadata)}}}'
        )
        assert (
            res.dict()
            == res.dict()
            == {
                "tag": "payRequest",
                "callback": "https://service.io/pay",
                "minSendable": 1000,
                "maxSendable": 2000,
                "metadata": metadata,
            }
        )

    @pytest.mark.parametrize(
        "d",
        [
            {"callback": "invalid", "minSendable": 1000, "maxSendable": 2000, "metadata": metadata},
            {"callback": "https://service.io/pay"},  # missing fields
            {"callback": "https://service.io/pay", "minSendable": 0, "maxSendable": 0, "metadata": metadata},  # 0
            {"callback": "https://service.io/pay", "minSendable": 100, "maxSendable": 10, "metadata": metadata},  # max
            {"callback": "https://service.io/pay", "minSendable": -90, "maxSendable": -10, "metadata": metadata},
        ],
    )
    def test_invalid_data(self, d):
        with pytest.raises(ValidationError):
            LnurlPayResponse(**d)


class TestLnurlPayResponseComment:
    @pytest.mark.parametrize(
        "callback, min_sendable, max_sendable, metadata, comment_allowed",
        [
            (
                "https://service.io/pay",
                1000,
                2000,
                metadata,
                555,  # comment allowed
            ),
        ],
    )
    def test_success_response(
        self, callback: str, min_sendable: int, max_sendable: int, metadata: str, comment_allowed: int
    ):
        res = LnurlPayResponse(
            callback=parse_obj_as(CallbackUrl, callback),
            minSendable=MilliSatoshi(min_sendable),
            maxSendable=MilliSatoshi(max_sendable),
            metadata=parse_obj_as(LnurlPayMetadata, metadata),
            commentAllowed=comment_allowed,
        )
        assert res.ok
        assert (
            res.json() == res.json() == '{"tag": "payRequest", "callback": "https://service.io/pay", '
            f'"minSendable": 1000, "maxSendable": 2000, "metadata": {json.dumps(metadata)}, '
            '"commentAllowed": 555}'
        )
        assert (
            res.dict()
            == res.dict()
            == {
                "tag": "payRequest",
                "callback": "https://service.io/pay",
                "minSendable": 1000,
                "maxSendable": 2000,
                "metadata": metadata,
                "commentAllowed": 555,
            }
        )

    @pytest.mark.parametrize(
        "d",
        [
            {"callback": "invalid", "minSendable": 1000, "maxSendable": 2000, "metadata": metadata},
            {"callback": "https://service.io/pay"},  # missing fields
            {"callback": "https://service.io/pay", "minSendable": 0, "maxSendable": 0, "metadata": metadata},  # 0
            {"callback": "https://service.io/pay", "minSendable": 100, "maxSendable": 10, "metadata": metadata},  # max
            {"callback": "https://service.io/pay", "minSendable": -90, "maxSendable": -10, "metadata": metadata},
            {
                "callback": "https://service.io/pay",
                "minSendable": 100,
                "maxSendable": 1000,
                "metadata": metadata,
                "commentAllowed": "Yes",  # str should be int
            },
        ],
    )
    def test_invalid_data(self, d):
        with pytest.raises(ValidationError):
            LnurlPayResponse(**d)


class TestLnurlWithdrawResponse:
    @pytest.mark.parametrize(
        "callback, k1, min_withdrawable, max_withdrawable",
        [
            (
                "https://service.io/withdraw",
                "c3RyaW5n",
                100,
                200,
            ),
        ],
    )
    def test_success_response(self, callback: str, k1: str, min_withdrawable: int, max_withdrawable: int):
        res = LnurlWithdrawResponse(
            callback=parse_obj_as(CallbackUrl, callback),
            k1=k1,
            minWithdrawable=MilliSatoshi(min_withdrawable),
            maxWithdrawable=MilliSatoshi(max_withdrawable),
        )
        assert res.ok
        assert (
            res.json()
            == res.json()
            == '{"tag": "withdrawRequest", "callback": "https://service.io/withdraw", "k1": "c3RyaW5n", '
            '"minWithdrawable": 100, "maxWithdrawable": 200, "defaultDescription": ""}'
        )
        assert (
            res.dict()
            == res.dict()
            == {
                "tag": "withdrawRequest",
                "callback": "https://service.io/withdraw",
                "k1": "c3RyaW5n",
                "minWithdrawable": 100,
                "maxWithdrawable": 200,
                "defaultDescription": "",
            }
        )

    @pytest.mark.parametrize(
        "d",
        [
            {"callback": "invalid", "k1": "c3RyaW5n", "minWithdrawable": 1000, "maxWithdrawable": 2000},
            {"callback": "https://service.io/withdraw", "k1": "c3RyaW5n"},  # missing fields
            {
                "callback": "https://service.io/withdraw",
                "k1": "c3RyaW5n",
                "minWithdrawable": 100,
                "maxWithdrawable": 10,
            },
            {"callback": "https://service.io/withdraw", "k1": "c3RyaW5n", "minWithdrawable": -9, "maxWithdrawable": -1},
        ],
    )
    def test_invalid_data(self, d):
        with pytest.raises(ValidationError):
            LnurlWithdrawResponse(**d)

    @pytest.mark.parametrize(
        "payLink",
        [
            "https://service.io/withdraw",
            "lnurlw://service.io/withdraw",  # wrong LUD17
            str(encode("https://service.io/withdraw").bech32),  # bech32
        ],
    )
    def test_invalid_pay_link(self, payLink: str):
        with pytest.raises(ValidationError):
            _ = LnurlWithdrawResponse(
                callback=parse_obj_as(CallbackUrl, "https://service.io/withdraw/cb"),
                k1="c3RyaW5n",
                minWithdrawable=MilliSatoshi(100),
                maxWithdrawable=MilliSatoshi(200),
                payLink=payLink,
            )

    def test_valid_pay_link(self):
        payLink = parse_obj_as(Lnurl, "lnurlp://service.io/pay")
        assert payLink.is_lud17
        assert payLink.lud17_prefix == "lnurlp"
        _ = LnurlWithdrawResponse(
            callback=parse_obj_as(CallbackUrl, "https://service.io/withdraw/cb"),
            k1="c3RyaW5n",
            minWithdrawable=MilliSatoshi(100),
            maxWithdrawable=MilliSatoshi(200),
            payLink=payLink.lud17,
        )
